//=============================================================================
/// @class DTime dutil/DTime.h
/// @brief encapsulates system time
/// @author Daniel Larimer
//=============================================================================
#ifndef _DTIME_H_
#define _DTIME_H_

#ifdef WIN32
#include <time.h>
#include <sys/timeb.h>
#include <winsock2.h>

#ifdef DUTIL_EXPORTS
#define PEXPR_API __declspec(dllexport)
#else
#define PEXPR_API __declspec(dllimport)
#endif /* DUTIL_EXPORTS */

#ifdef __cplusplus
extern "C" {
#endif

inline PEXPR_API int gettimeofday( timeval* t, int )
{
	_timeb wintime;
	_ftime( &wintime );
	t->tv_sec = wintime.time;
	t->tv_usec = wintime.millitm * 1000;
	return 0;
};

#ifdef __cplusplus
}
#endif

#else
#define PEXPR_API
#endif /* WIN32 */


class PEXPR_API DTime
{
    public:
        DTime( int usec = 0 );
        DTime( int sec, int usec );
        DTime( int min, int sec, int usec );
        DTime( int hr, int min, int sec, int usec );
        static DTime current();

        void  start();
        DTime reset();
        DTime elapsed()const;
       
        float hour()const   { return minute()/60.; }
        float minute()const { return sec()/60.;    }
        float sec()const    { return msec()/1000.; }
        float msec()const   { return m_usec/1000.; }
        int   usec()const   { return m_usec; }

        DTime  operator- (const DTime& b ) { return DTime( m_usec - b.m_usec ); }
        DTime& operator-=(const DTime& b ) { m_usec -= b.m_usec; return *this;  }
        DTime  operator+ (const DTime& b ) { return DTime( m_usec + b.m_usec ); }
        DTime& operator+=(const DTime& b ) { m_usec += b.m_usec; return *this;  }
        bool operator> (const DTime& b ) { return m_usec >  b.m_usec; }
        bool operator>=(const DTime& b ) { return m_usec >= b.m_usec; }
        bool operator< (const DTime& b ) { return m_usec <  b.m_usec; }
        bool operator<=(const DTime& b ) { return m_usec <= b.m_usec; }
        bool operator==(const DTime& b ) { return m_usec == b.m_usec; }
        bool operator!=(const DTime& b ) { return m_usec != b.m_usec; }

    private:
		int   m_usec; // micro-sec from Jan 1, 1970
        int   m_start;
};

#endif
