//=============================================================================
// POBox.h
/// @class POBox POBox.h
/// @author Daniel Larimer
/// @brief - a receiving queue for messages
//
// Created 5/13/03 By Daniel Larimer
//=============================================================================
#ifndef _POBOX_H_
#define _POBOX_H_
#include<qobject.h>
#include<string>
#include<list>
#include<DTime.h>
#include<POAddress.h>
#include<POBoxIF.h>

using namespace std;
class PostOffice;
class POMessage;

class PEXPR_API POBox : public QObject, public POBoxIF
{
    Q_OBJECT
    public:
        POBox( const string& name, PostOffice* po );
        ~POBox();

        const POMessage* popMessage();
        
        void   setQueueState( bool state = true )   { m_queuestate = state; }
        bool   getQueueState()const                 { return m_queuestate;  }

        string getName()const { return m_name; }
        bool   hasMessages( bool check_now = false )const;

        PostOffice* getPostOffice()const{ return m_po; }
    
        bool sendTextMsg( const POAddress& adr, const char* format, ... );
        bool sendTextMsg( const string& adr, const char* format, ... );
        
    
        bool sendMessage( POMessage* msg, const POAddress& adr, 
                                        bool send_now = false );
        const POMessage* sendBlockingMessage( POMessage* msg, 
                                        const POAddress& adr, 
                                        const DTime& timeout = DTime(2,0) );

        void joinGroup( const POAddress& groupadr, bool create = false );
        void joinGroup( const string& groupname, bool create = false );
        void leaveGroup( const POAddress& groupadr );
        void leaveGroup( const string& groupname );
    
    public slots:
        void addMessage( const POMessage* msg );

    signals:
        /// emited when a new message arives if QueueState is true
        /// slots hooked to this signal shouldn't set the QueueState to false
        /// unless they want the POBox to delete the message for them
        void messageArrived();

        /// emited for every message that arrives
        void messageArrived( const POMessage* msg );
        void messageSent   ( const POMessage* msg );
        
    private:
        list<const POMessage*>    m_inbox;
        string                    m_name;
        PostOffice*               m_po;
        bool                      m_queuestate;
};

#endif
