#include"POClient.h"
#include"POBoxListMsg.h"
#include"POPartialMsg.h"
#include"POErrorMsg.h"
#include"POMessageFactory.h"
#include<iostream>
#include<qsocket.h>
#include<qapplication.h>
#include<Tracer.h>

#define BUFFERSIZE (1024*1024*5) //5MB

using namespace std;

//=============================================================================
// Function Name: POClient
///     Joins an existing socket
/// \param fd - the socket descriptor
//=============================================================================
POClient::POClient( int fd )
{
    m_buffer = new char[BUFFERSIZE];
    m_usedsize = 0;
    m_incommingsize = 0;

    m_socket = new QSocket();
    m_socket->setSocket( fd );

    m_peerport = m_socket->peerPort();

    m_state = WAIT_CLIST;
    m_clientname = "undefined";

    connect( m_socket, SIGNAL( readyRead() ), 
             this,     SLOT  ( read()      ) );
    connect( m_socket, SIGNAL( connectionClosed() ),
             this,     SLOT  ( socketClosed() ) );
    connect( m_socket, SIGNAL( error(int) ),
             this,     SLOT  ( error(int) ) );

    //Tracer::debug( "Creating new POClient" );
}

//=============================================================================
// Function Name: POClient
///     Opens a new socket to the specified host and port, when the socket is
///     connected SocketConnected() will be emited at which point the server
///     will be waiting for a POBoxListMsg from the local PostOffice
/// \param host - the host to connect to
/// \param port - the port
/// \return 
//=============================================================================
POClient::POClient( const string& host, int port )
{
    m_buffer = new char[BUFFERSIZE];
    m_usedsize = 0;
    m_incommingsize = 0;

    m_peeradr  = host;
    m_peerport = port;

    m_socket = new QSocket();
    m_socket->connectToHost( host.c_str(), port );


    m_state = WAIT_CLIST;
    m_clientname = "undefined";
    
    connect( m_socket, SIGNAL( readyRead() ), 
             this,     SLOT  ( read()      ) );
    connect( m_socket, SIGNAL( connectionClosed() ),
             this,     SLOT  ( socketClosed() ) );
    connect( m_socket, SIGNAL( connected() ),
             this,     SLOT  ( socketConnected() ) );
    connect( m_socket, SIGNAL( error(int) ),
             this,     SLOT  ( error(int) ) );
}
POClient::~POClient()
{
    fprintf( stderr, "~POClient\n" );
    delete[] m_buffer;
    delete m_socket;
}
//=============================================================================
// Function Name: SocketConnected
///  called when the socket makes a connection and emits SocketConnected(this)
//=============================================================================
void POClient::socketConnected()
{
    emit socketConnected( this );
}

//=============================================================================
// Function Name: Read
///  called when there is data to be read.  Reads the data loads it into the
///  buffer, searches for SYNC and builds a message buffer
//=============================================================================
void POClient::read()
{
    if( m_socket->bytesAvailable() !=  0)
    {
        int readsize = m_socket->bytesAvailable();
        if( readsize + m_usedsize > BUFFERSIZE )
            readsize = BUFFERSIZE - m_usedsize;
        
        //====================================
        // Read The Socket
        //====================================
        m_usedsize += m_socket->readBlock( m_buffer+m_usedsize, readsize );
        //====================================
    }

    //==========================================
    // Parse Input Buffer for Complete Message
    //==========================================
    if( m_incommingsize <= 0 )
    {
        //Tracer::debug("POClient::read() - m_incomming size <= 0");
        // Trim everything until the first SYNC
        string sync("SYNC");
        for( int i = 0; i < m_usedsize - 4; i++ )
        {
            if(sync == string(&m_buffer[i], 4))
            {
                m_usedsize = m_usedsize - i - 4;
                memmove( m_buffer, m_buffer + i + 4, m_usedsize );
                break;
            }
        }
        // If there is anything left in the buffer, it is the start 
        // of a message
        if( m_usedsize > 0 )
        {
            if( (unsigned)m_usedsize < sizeof(int) ) // we need to wait on the message size
            {
                m_incommingsize = 4;
            }
            m_incommingsize.unpack(m_buffer);
            if( m_incommingsize < 0 ) m_incommingsize = 0;
        }
    }
    //==================================================
    // Are we waiting on the message size?
    //==================================================
    if( m_incommingsize == 4 && m_usedsize >= 4 )
    {
        m_incommingsize.unpack(m_buffer);
        if( m_incommingsize < 0 ) m_incommingsize = 0;
    }else if( m_usedsize >= m_incommingsize )
    {
        //Build the Message and deliver it
        POMessage* msg = NULL;
        try {
            msg = POMessageFactory::BuildFromBuffer( m_buffer );
        }catch( POErrorMsg* err )
        {
            Tracer::error( err->getText().c_str() );
            m_usedsize -= m_incommingsize;
            memmove( m_buffer, m_buffer+m_incommingsize, m_usedsize );
            m_incommingsize = 0;
//            throw err; // this will crash Qt because read is called from QApplication
        }
// The factory now throws an exception 
//        if( msg == NULL )
//        {
//            exit(1);
//        }
//        Tracer::debug("POClient::read() - message arrived: %s\n",
//                      msg->messageType().c_str() );
        
        m_usedsize -= m_incommingsize;
        memmove( m_buffer, m_buffer+m_incommingsize, m_usedsize );
        m_incommingsize = 0;
        
        if( msg->messageType() == "POBoxListMsg" )
            handleBoxListMsg( (POBoxListMsg*)msg );
        else if( msg->messageType() == "POPartialMsg" )
            handleSegment( (POPartialMsg*)msg );
        else
        {
            emit receivedMessage( msg );
            delete msg;
        }
        
        if( m_usedsize > 0 )
            read();
    }
}

void POClient::handleSegment( POPartialMsg* msg )
{
    m_msgsegments.push_back(msg);
    
    // see if this was the last in the series
    if( msg->getNumber() == msg->getMaxNum() )
    {
        // calculate total size
        int size = 0;
        for( unsigned int i = 0; i < m_msgsegments.size(); i++ )
            size += m_msgsegments[i]->getBufferSize();
        // create the new msg buffer
        char* buffer = new char[size];
        char* tmp = buffer;
        // copy the segments into the new buffer
        for( unsigned int x = 0; x < m_msgsegments.size(); x++ )
        {
            memcpy( tmp, m_msgsegments[x]->getData(), 
                    m_msgsegments[x]->getBufferSize() );
            tmp += m_msgsegments[x]->getBufferSize();
            delete m_msgsegments[x]; // we no longer need it
        }
        m_msgsegments.clear();

        // unpack the big message
        POMessage* msg = NULL;
        try
        {
            msg = POMessageFactory::BuildFromBuffer( m_buffer );
        }
        catch( POErrorMsg* err  )
        {
            Tracer::error( err->getText().c_str() ); 
            delete[] buffer;
            delete msg;
//            throw err;  // if we throw this it will be caught in QApplication and crash
        }
        // Do some clean up before more cloning
        delete[] buffer;
        emit receivedMessage( msg );
        delete   msg;
    }
}

void POClient::socketClosed()
{
    m_state = DISCONNECTED;
    emit connectionLost( this );
    Tracer::stat( "Client Closed" );
}

//=============================================================================
// Function Name: sendMessasge
///     Assumes that the message->from field has been set and then sees if
///     the remote PostOffice has a mailbox for adr->pobox;
/// \throw POErrorMsg* if the message could not be sent
/// \param adr - the destination address
//=============================================================================
void POClient::sendMessage( const POAddress& adr, const POMessage* msg )
{
    // Verify that address mailbox is found on server
    for( unsigned int i = 0; i < m_clientboxes.size(); i++ )
        if( m_clientboxes[i] == adr.getBox() )
        {
            sendMessage( msg );
            return;
        }
    throw new POErrorMsg( "POBox \"%s\" not found at Post Office \"%s\".",
                           adr.getPO().c_str(), m_clientname.c_str() );
}

//=============================================================================
// Function Name: sendMessage
///     Sends the message regardless of destination
/// \param msg - the message to send
//=============================================================================
void POClient::sendMessage( const POMessage* msg )
{
    char* buffer = new char[msg->packsize()+4];
    memcpy( buffer, "SYNC", 4 );
    char* tmp = buffer + 4;
    long size = msg->pack(tmp);
    if( size != msg->packsize() )
    {
        Tracer::error( "POClient::sendMessage() pack() != packsize()\n"
                       "\tMessageType = %s\n"
                       "\tpack() = %d vs. packsize() = %d", 
                       msg->messageType().c_str(), size, msg->packsize() );
        exit(1);
    }
    size += 4;
    if( size > MAX_MSG_SIZE )
    {
        char* mtmp = tmp;
        int nummsgs = size / (MAX_MSG_SIZE-1024); // save a k for overhead
        // check to see if it came out evenly
        if( float(size) - (nummsgs*(MAX_MSG_SIZE-1024) ) > 0 )
            nummsgs += 1; // left over
        int remaining = size;
        cout<<"nummsgs = "<<nummsgs<<endl;
        int count = 0;
        while( remaining > 0 )
        {
            POPartialMsg* pmsg;
            if( remaining >= (MAX_MSG_SIZE-1024) )
            {
                pmsg = new POPartialMsg( count, nummsgs, 
                                          mtmp, (MAX_MSG_SIZE-1024) );
                remaining -= (MAX_MSG_SIZE-1024);
            }else{
                pmsg = new POPartialMsg( count, nummsgs, mtmp, remaining );
                remaining = 0;
            }
            count++;
            sendMessage( pmsg );
            delete pmsg;
        }
        cout<<"nummsgs = "<<nummsgs<<endl;
        cout<<"send nummsgs = "<<count<<endl;
    }
    m_socket->writeBlock( buffer, size );
    delete[] buffer;
}

void POClient::handleBoxListMsg( const POBoxListMsg* msg )
{
    m_clientname = msg->getSource().getPO();
    vector<string> boxes = msg->getBoxes();
    if( msg->getCommand() == POBoxListMsg::NEWBOXES )
    {
        for( unsigned int i = 0; i < boxes.size(); i++ )
        {
            // TBD: Make sure the box is unique
            m_clientboxes.push_back( boxes[i] );
         //   Tracer::stat( "Client %s registered mailbox: %s",
         //                 msg->getSource().getPO().c_str(),
         //                 boxes[i].c_str() );
            emit newPOBox( m_clientname, boxes[i] );
        }
    }else{
        for( unsigned int i = 0; i < boxes.size(); i++ )
            removeClientBox( boxes[i] );
    }
    delete msg;
    if( m_state != CONNECTED )
    {
        m_state = CONNECTED;
        emit connectionEstablished( this );
    }
}

//=============================================================================
// Function Name: RemoveClientBox
///     Removes a box from the list of remote boxes
/// \param box - the box to remove
//=============================================================================
void POClient::removeClientBox( const string& box )
{
    vector<string>::iterator itr = m_clientboxes.begin();
    while( itr != m_clientboxes.end() )
    {
        if( *itr == box )
        {
            m_clientboxes.erase(itr);
            return;
        }
        itr++;
    }
    //Tracer::warn("Couldn't remove remote client box: %s", box.c_str() );
}

void POClient::error( int err )
{
    switch( err )
    {
        case QSocket::ErrConnectionRefused:
            Tracer::error( "Connection to host: %s on port: %d "
                           "was refused.", m_peeradr.c_str(), m_peerport );
//            cerr<<"Connection to host: "<<m_peeradr<<" on port: "
//               <<m_peerport<<" was refused\n";
            break;
        case QSocket::ErrHostNotFound:
            Tracer::error( "Host %s not found.", m_peeradr.c_str() );
//            cerr<<"Host "<<m_peeradr<<" not found.\n";
            break;
        case QSocket::ErrSocketRead:
            Tracer::error( "Error reading socket");
//            cerr<<"Error reading socket\n";
            break;
    }
    emit connectionLost( this );
}
