//=============================================================================
// POClient.h
//
// Created 5/13/03 By Daniel Larimer
//=============================================================================
#ifndef _POCLIENT_H_
#define _POCLIENT_H_
#include<qobject.h>
#include<string>
#include<vector>
#include<qstring.h>
#include<POAddress.h>

// the maximum size of a message before
// it is broken up into multiple messages
#define MAX_MSG_SIZE (1024*1024*4) // 4MB

using namespace std;

class QSocket;
class POMessage;
class POBoxListMsg;
class POPartialMsg;

class PEXPR_API POClient : public QObject
{
    Q_OBJECT
    public:
        enum CLIENT_STATE { CONNECTED, WAIT_CLIST, DISCONNECTED };
        POClient( int fd );
        POClient( const string& host, int port );
        ~POClient();
        
        string getClientName()const { return m_clientname; }
        int    getState()const { return m_state; }
    
        vector<string>  getAllBoxes()const { return m_clientboxes; }
    public slots:
        void sendMessage( const POAddress& adr, const POMessage* msg );
        void sendMessage( const POMessage* msg );

    private slots:
        void error( int );
        void socketConnected();
        void socketClosed();
        void read();
        
    signals:
        void socketConnected( POClient* client );
        void connectionEstablished( POClient* client );
        void receivedMessage( const POMessage* msg );
        void connectionLost( POClient* );
        void newPOBox( const string& clientname, const string& boxname );
    
    private:
        void handleSegment( POPartialMsg* msg );
        void handleBoxListMsg( const POBoxListMsg* msg );
        void removeClientBox( const string& box );
    
        int             m_state;
        string          m_clientname;
        vector<string>  m_clientboxes;
        QSocket*        m_socket;
        char*           m_buffer;
        int             m_usedsize;
        string          m_peeradr;
        int             m_peerport;

        POInt                   m_incommingsize;
        vector<POPartialMsg*>   m_msgsegments;
};

#endif
