#include"POErrorMsg.h"
#include"POMessageFactory.h"
#include<stdarg.h>

//=============================================================================
// Function Name: POErrorMsg
///     Default constructor
//=============================================================================
POErrorMsg::POErrorMsg( const string& txt )
:m_txt(txt)
{
}

POErrorMsg::POErrorMsg( const POErrorMsg& cpy )
:POMessage((POMessage)cpy), m_txt(cpy.m_txt)
{
}
POErrorMsg::POErrorMsg( const char* format, ... )
{
    char* temp = new char[5000];
    va_list ap;
    va_start( ap, format );
    vsnprintf( temp, 4999, format, ap );
    va_end(ap);
    temp[4999] = '\0';
    m_txt = string(temp);
}
POErrorMsg::~POErrorMsg()
{
}

void POErrorMsg::setText( const string& txt )
{
    m_txt = txt;
}


//=============================================================================
// Function Name: clone
///     Creates an identical copy of this message.  This function must be 
///     overloaded in derived classes.
/// \return a copy of this message
//=============================================================================
POMessage* POErrorMsg::clone()const
{
    return new POErrorMsg( *this );
}

//=============================================================================
// Function Name: packsize
///     calculates the number of bytes needed to pack this message
/// \return size in bytes
//=============================================================================
int POErrorMsg::packsize()const
{
    int size = POMessage::packsize();
    size += m_txt.packsize();
    return size;
}

//=============================================================================
// Function Name: pack
///     packs the message into a pre-alocated buffer of size >= packsize()
/// \param buffer - the buffer to pack into
/// \return the number of bytes packed
//=============================================================================
int POErrorMsg::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += POMessage::pack(tmp);
    tmp += m_txt.pack(tmp); 
    
    return tmp - buffer;
}

//=============================================================================
// Function Name: unpack
///     unpacks the message, setting source, destination, and derived type
/// \param buffer the buffer to unpack, must be valid, no error checking
/// \return the number of bytes unpacked
//=============================================================================
int POErrorMsg::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += POMessage::unpack(tmp);
   
    tmp += m_txt.unpack( tmp );

    return tmp - buffer;
}

