#include"POGroup.h"
#include<PostOffice.h>
#include<POBox.h>
#include<POErrorMsg.h>
#include<POMessage.h>
#include<POGroupMsg.h>
#include<Tracer.h>

//=============================================================================
// Function Name: POGroup
///     Constructor
/// \param name - the name of the group
/// \param po - the post office the group belongs to
//=============================================================================
POGroup::POGroup( const string& name, PostOffice* po )
{
    m_po = po;
    m_inbox = m_po->registerPOBox( name );
    connect( m_inbox, SIGNAL( messageArrived( const POMessage* ) ),
             this,    SLOT  ( handleMessage( const POMessage* ) ) );
}

//=============================================================================
// Function Name: ~POGroup
///     Destructor, closes the POBox
//=============================================================================
POGroup::~POGroup()
{
    // TODO: Notify All Members
    m_po->closePOBox( m_inbox->getName() );
}

//=============================================================================
// Function Name: handleMessage
///     Forwards all non GroupMsg messages to all members of the group
///     w/out changing the source of the message
/// \param msg - the message to respond to
//=============================================================================
void POGroup::handleMessage( const POMessage* msg )
{
    //======================================
    // Handle Group Messages
    //======================================
    if( msg->messageType() == "POGroupMsg" )
    {
        POGroupMsg* gmsg = (POGroupMsg*)msg;
        if( gmsg->getCommand() == POGroupMsg::JOIN )
            addAddress( msg->getSource() );
        else if( gmsg->getCommand() == POGroupMsg::LEAVE )
            removeAddress( msg->getSource() );
        else if( gmsg->getCommand() == POGroupMsg::ADD )
            addAddress( gmsg->getAddress() );
        else if( gmsg->getCommand() == POGroupMsg::REMOVE )
            removeAddress( gmsg->getAddress() );
    }
    else // forward the message
        forward(msg);
}

void POGroup::addAddress( const POAddress& adr )
{
    //===================================
    // Verify that the address is unique
    //===================================
    for( unsigned int i = 0; i < m_members.size(); i++ )
        if( m_members[i] == adr )
            return; // address already in group
    m_members.push_back( adr );
}

void POGroup::removeAddress( const POAddress& adr )
{
     vector<POAddress>::iterator itr = m_members.begin();
     while( itr != m_members.end() )
     {
         if( *itr == adr )
         {
             m_members.erase(itr);
             return;
         }
         itr++;
     }
}

void POGroup::forward( const POMessage* msg )
{
    POMessage* cpy = msg->clone();
    for( unsigned int i = 0; i < m_members.size(); i++ )
    {
        if( msg->getSource() != m_members[i] )
        {   
            cpy->setDestination( m_members[i] );
            try{
                m_po->sendMessage( cpy );
            }catch ( POErrorMsg* err )
            {
                Tracer::warn( "Removing %s.%s from group %s.%s",
                              m_members[i].getPO().c_str(),
                              m_members[i].getBox().c_str(),
                              m_inbox->getPostOffice()->getName().c_str(),
                              m_inbox->getName().c_str() );
                m_members.erase( m_members.begin() + i );
                i--;
                delete err;
            }
        }
    }
    delete cpy;
}
