#include"POGroupMsg.h"
#include"POMessageFactory.h"

#include<iostream>
using namespace std;

//=============================================================================
// Function Name: POGroupMsg
///     Default constructor
//=============================================================================
POGroupMsg::POGroupMsg( int cmd )
:m_cmd(cmd)
{
}

//=============================================================================
// Function Name: POGroupMsg
///     Constructs a message that will add or remove an address from a group
/// \param cmd - ADD || REMOVE
/// \param adr - the address to add/remove to/from the group
/// \return 
//=============================================================================
POGroupMsg::POGroupMsg( const POAddress& adr, int cmd )
:m_cmd(cmd), m_adr(adr)
{
}

/// Copy Constructor
POGroupMsg::POGroupMsg( const POGroupMsg& cpy )
:POMessage((POMessage)cpy), m_cmd(cpy.m_cmd), m_adr(cpy.m_adr)
{
}

POGroupMsg::~POGroupMsg()
{
}

//=============================================================================
// Function Name: clone
///     Creates an identical copy of this message.  This function must be 
///     overloaded in derived classes.
/// \return a copy of this message
//=============================================================================
POMessage* POGroupMsg::clone()const
{
    return new POGroupMsg( *this );
}

//=============================================================================
// Function Name: packsize
///     calculates the number of bytes needed to pack this message
/// \return size in bytes
//=============================================================================
int POGroupMsg::packsize()const
{
    int size = POMessage::packsize();
    size += m_cmd.packsize();
    if( m_cmd.val() == ADD || m_cmd.val() == REMOVE )
        size += m_adr.packsize();
    return size;
}

//=============================================================================
// Function Name: pack
///     packs the message into a pre-alocated buffer of size >= packsize()
/// \param buffer - the buffer to pack into
/// \return the number of bytes packed
//=============================================================================
int POGroupMsg::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += POMessage::pack(tmp);
    tmp += m_cmd.pack(tmp); 
    if( m_cmd.val() == ADD || m_cmd.val() == REMOVE )
        tmp += m_adr.pack(tmp);
    
    return tmp - buffer;
}

//=============================================================================
// Function Name: unpack
///     unpacks the message, setting source, destination, and derived type
/// \param buffer the buffer to unpack, must be valid, no error checking
/// \return the number of bytes unpacked
//=============================================================================
int POGroupMsg::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += POMessage::unpack(tmp);
   
    tmp += m_cmd.unpack( tmp );
    if( m_cmd.val() == ADD || m_cmd.val() == REMOVE )
        tmp += m_adr.unpack(tmp);

    return tmp - buffer;
}

