#include"POMessage.h"
#include"POMessageFactory.h"
#include<iostream>
#include<Tracer.h>

using namespace std;

//=============================================================================
// Function Name: POMessage
///     Default constructor
//=============================================================================
POMessage::POMessage( )
:m_derivedtype("POMessage")
{
}
POMessage::POMessage( const POMessage& cpy )
:Packable(),
 m_source( cpy.m_source ), 
 m_destination( cpy.m_destination ),
 m_derivedtype(cpy.m_derivedtype),
 m_tag(cpy.m_tag)
{
    
}

POMessage::~POMessage()
{
}

POAddress POMessage::getSource()const
{
    return m_source;
}

POAddress POMessage::getDestination()const
{
    return m_destination;
}   

void POMessage::setSource( const POAddress& source )
{
    m_source = source;
}

void POMessage::setDestination( const POAddress& dest )
{
    m_destination = dest;
}


//=============================================================================
// Function Name: clone
///     Creates an identical copy of this message.  This function must be 
///     overloaded in derived classes.
/// \return a copy of this message
//=============================================================================
POMessage* POMessage::clone()const
{
    return new POMessage( *this );
}

//=============================================================================
// Function Name: packsize
///     calculates the number of bytes needed to pack this message
/// \return size in bytes
//=============================================================================
int POMessage::packsize()const
{
    int size = 0;
    size += sizeof(int); // message size
    size += POString( messageType() ).packsize();
    size += m_source.packsize();
    size += m_destination.packsize();
    size += m_tag.packsize();
    return size;
}

//=============================================================================
// Function Name: pack
///     packs the message into a pre-alocated buffer of size >= packsize()
/// \param buffer - the buffer to pack into
/// \return the number of bytes packed
//=============================================================================
int POMessage::pack( char* buffer )const
{
    char* tmp  = buffer;
    
    POInt size(packsize());
    tmp += size.pack(tmp);

    POString type(messageType());
    tmp += type.pack( tmp );
    tmp += m_source.pack( tmp );
    tmp += m_destination.pack( tmp );
    tmp += m_tag.pack(tmp);
    return tmp - buffer;
}

//=============================================================================
// Function Name: unpack
///     unpacks the message, setting source, destination, and derived type
/// \param buffer the buffer to unpack, must be valid, no error checking
/// \return the number of bytes unpacked
//=============================================================================
int POMessage::unpack( char* buffer )
{
    char* tmp = buffer;
   
    POInt size;
    tmp += size.unpack(tmp);
    
    tmp += m_derivedtype.unpack( tmp );
    tmp += m_source.unpack( tmp );
    tmp += m_destination.unpack( tmp );
    tmp += m_tag.unpack(tmp);

    return tmp - buffer;
}

