#include"POPartialMsg.h"
#include"POMessageFactory.h"


//=============================================================================
// Function Name: POPartialMsg
/// @param num - the number in series out of max
/// @param of - the number of segments in the series
/// @param data - the data to send
/// @param size - the size of the data
//=============================================================================
POPartialMsg::POPartialMsg( int num, int of, char* data, int size )
:POMessage()
{
    m_num = num;
    m_maxnum = of;
    m_size = size;
    if( size <= 0 )
    {
        m_buffer = NULL;
        return;
    }
    m_buffer = new char[size];
    memcpy( m_buffer, data, size );
}

POPartialMsg::POPartialMsg( const POPartialMsg& cpy )
:POMessage((POMessage)cpy)
{
    m_size = cpy.m_size;
    m_num  = cpy.m_num;
    m_maxnum = cpy.m_maxnum;
    m_buffer = NULL;
    if( m_size )
    {
        m_buffer = new char[m_size.val()];
        memcpy( m_buffer, cpy.m_buffer, m_size.val() );
    }
}

POPartialMsg::~POPartialMsg()
{
    if( m_buffer )
        delete[] m_buffer;
}


//=============================================================================
// Function Name: clone
///     Creates an identical copy of this message.  This function must be 
///     overloaded in derived classes.
/// \return a copy of this message
//=============================================================================
POMessage* POPartialMsg::clone()const
{
    return new POPartialMsg( *this );
}

//=============================================================================
// Function Name: packsize
///     calculates the number of bytes needed to pack this message
/// \return size in bytes
//=============================================================================
int POPartialMsg::packsize()const
{
    int size = POMessage::packsize();
    size += m_size.packsize();
    size += m_num.packsize();
    size += m_maxnum.packsize();
    size += m_size.val();
    return size;
}

//=============================================================================
// Function Name: pack
///     packs the message into a pre-alocated buffer of size >= packsize()
/// \param buffer - the buffer to pack into
/// \return the number of bytes packed
//=============================================================================
int POPartialMsg::pack( char* buffer )const
{
    char* tmp = buffer;
    tmp += POMessage::pack(tmp);
    tmp += m_size.pack(tmp);
    tmp += m_num.pack(tmp);
    tmp += m_maxnum.pack(tmp);

    if( m_size.val() > 0 )
    {
        memcpy( tmp, m_buffer, m_size.val() );
        tmp += m_size.val();
    }
    
    return tmp - buffer;
}

//=============================================================================
// Function Name: unpack
///     unpacks the message, setting source, destination, and derived type
/// \param buffer the buffer to unpack, must be valid, no error checking
/// \return the number of bytes unpacked
//=============================================================================
int POPartialMsg::unpack( char* buffer )
{
    char* tmp = buffer;
    tmp += POMessage::unpack(tmp);
    tmp += m_size.unpack(tmp);
    tmp += m_num.unpack(tmp);
    tmp += m_maxnum.unpack(tmp);
   
    if( m_size.val() > 0 )
    {
        if( m_buffer ) delete[] m_buffer;
        m_buffer = new char[m_size.val()];
        memcpy( m_buffer, tmp, m_size.val() );
        tmp += m_size.val();
    }

    return tmp - buffer;
}

