#include"POSmallFileMsg.h"
#include<Tracer.h>
#include<qfile.h> // Qt is required by pony express anyway

POSmallFileMsg::POSmallFileMsg( const string& srcname,
                                const string& destfile)
    :POMessage()
{
    m_filedata = NULL;
    m_filesize = 0;
    loadFile( srcname );
    m_destfile = destfile;
}

POSmallFileMsg::POSmallFileMsg( const POSmallFileMsg& cpy )
    :POMessage( cpy )
{
    m_filesize = cpy.m_filesize;
    m_filename = cpy.m_filename;
    m_destfile = cpy.m_destfile;
    if( m_filesize.val() > 0 )
    {
        m_filedata = new char[m_filesize.val()];
        memcpy( m_filedata, cpy.m_filedata, m_filesize.val() );
    }
    else
        m_filedata = NULL;
}
POSmallFileMsg::~POSmallFileMsg()
{
    if( m_filedata ) delete[] m_filedata;
}
POMessage* POSmallFileMsg::clone()const
{
    return (POMessage*) new POSmallFileMsg( *this );
}   

bool POSmallFileMsg::loadFile( const string& fname )
{
    if( m_filedata ) delete[] m_filedata;
    m_filesize = 0;
    
    if( fname == "" )return false;

    if( !QFile::exists( fname.c_str() ) )
    {
        Tracer::error( "POSmallFileMsg::loadFile( %s )\n"
                       "\t\t- File doesn\'t exist." );
        return false;
    }
    QFile tmpfile( fname.c_str() );
    m_filesize = tmpfile.size();
    m_filedata = new char[m_filesize];
    tmpfile.readBlock( m_filedata, m_filesize );
    return true;
}

int POSmallFileMsg::packsize()const
{
    // always remember to pack the base class data
    int size = POMessage::packsize();
    size += m_filename.packsize(); // pack the file name
    size += m_destfile.packsize();
    size += m_filesize.packsize(); 
    size += m_filesize.val(); // the size of the file
    return size;
}

int POSmallFileMsg::pack( char* buffer )const
{
    char* tmp = buffer; // don't change the starting point
    tmp += POMessage::pack(tmp); // pack the base class 
                                 // and increment pointer 
    tmp += m_filename.pack(tmp); // pack the name
    tmp += m_destfile.pack(tmp); // the destination file
    tmp += m_filesize.pack(tmp); // pack the file size
     
    // return the number of bytes packed if filesize is 0
    if( m_filesize.val() <= 0 ) return tmp - buffer;
    
    // copy the file to the packed buffer
    memcpy( tmp, m_filedata, m_filesize.val() );
    tmp += m_filesize.val();
    return tmp - buffer; // tmp - buffer will be what you packed
}

// Be sure to unpack in the same order that you packed
int POSmallFileMsg::unpack( char* buffer )
{
    if( m_filedata ) delete[] m_filedata;
    m_filedata = NULL;
    
    char* tmp = buffer;
    tmp += POMessage::unpack(tmp);  // very easy to unpack
    tmp += m_filename.unpack(tmp);  // classes that implement
    tmp += m_destfile.unpack(tmp);
    tmp += m_filesize.unpack(tmp);  // this interface
    
    // if filesize is 0 there will be no file data
    if( m_filesize <= 0 ) return tmp - buffer; 
    
    // create the buffer to store the data
    m_filedata = new char[m_filesize];
    // copy the data
    memcpy( m_filedata, tmp, m_filesize );
    // increment the pointer
    tmp += m_filesize;
    
    return tmp - buffer; // the number of bytes unpacked
}



