#include"PostOffice.h"
#include"POClient.h"
#include"POServer.h"
#include"POBox.h"
#include"POMessage.h"
#include"POBoxListMsg.h"
#include<iostream>
#include<qdatetime.h>
#include<qapplication.h>
#include<Tracer.h>
#include<DTime.h>
#include<POGroup.h>
#include<POGroupMsg.h>
#include<POErrorMsg.h>
using namespace std;


map<string, PostOffice*>    PostOffice::m_postoffices;

//=============================================================================
// Function Name: PostOffice
///     Constructor
/// \param name - the name of the post office
//=============================================================================
PostOffice::PostOffice( const string& name )
:m_poname(name)
{
    if( m_postoffices.find(name) != m_postoffices.end() )
    {
        Tracer::warn("PostOffice::PostOffice( %s ) has already been created",
                     name.c_str() );
    }
    else
    {
        PostOffice::m_postoffices[name] = this;
    }
}

PostOffice::~PostOffice()
{
    cout<<"POstOffice;:~PostOffice() - CLEAN UP NEEDED\n";
}
//=============================================================================
// Function Name: getPOBox
///     returns a pointer to the specified pobox
/// \throws a POErrorMsg* if the POBox doesn't exist
/// \param name - the name for the POBox
/// \return a pointer to the new po box
//=============================================================================
POBox* PostOffice::getPOBox( const string& name )const
{
    cout<<"PostOffice::getPOBox( "<<name<<" )\n";
    map<string, POBox*>::const_iterator itr = m_localboxes.find(name);
    if( itr == m_localboxes.end() )
        throw new POErrorMsg( "PostOffice::getPOBox(\"%s\")\n"
                         "\tPOBox \"%s\" doesn't exist!\n",
                         name.c_str(), name.c_str() );
    return itr->second;
}

//=============================================================================
// Function Name: registerPOBox
///     Creates a new post office box and updates all connections
/// \throws a POErrorMsg* if the name has already been registered
/// \param name - the name for the POBox
/// \return a pointer to the new po box
//=============================================================================
POBox* PostOffice::registerPOBox( const string& name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(name);
    if( itr != m_localboxes.end() )
    {
        throw new POErrorMsg( "PostOffice::registerPOBox( \"%s\" )\n"
                              "\tPOBox with the name \"%s\" already registered.",
                              name.c_str(),
                              name.c_str()
                              );
        return NULL;
    }
    POBox* box = new POBox( name, this );
    m_localboxes[name] = box;

    POBoxListMsg msg( POBoxListMsg::NEWBOXES );
    msg.setSource( POAddress( m_poname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    msg.addBox( name );

    map<string, POClient*>::iterator cl = m_clients.begin();
    while( cl != m_clients.end() )
    {
        cl->second->sendMessage( &msg );        
        cl++;
    }
    emit newPOBox( m_poname, name );
    return box;
}

void PostOffice::newClientPOBox( const string& cname, const string& boxname )
{
    emit newPOBox( m_poname, cname, boxname );
}

//=============================================================================
// Function Name: ClosePOBox
///     Removes the box from the post officeThis
///     should only be called by ~POBox() when you delete it.
/// \param boxname - the name of the box to remove
/// \param dodelete - true if you want the post office box to be deleted as well
/// \return true if success, otherwise false
//=============================================================================
bool PostOffice::closePOBox( const string& boxname, bool dodelete )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(boxname);
    if( itr == m_localboxes.end() )
    {
        Tracer::error("PostOffice::closePOBox()\n"
                      "\t\tPOBox with the name \"%s\" isn't registered", 
                      boxname.c_str() );
        return false;
    }

    POBoxListMsg msg( POBoxListMsg::CLOSEBOXES );
    msg.setSource( POAddress( boxname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    msg.addBox( boxname );

    map<string, POClient*>::iterator cl = m_clients.begin();
    while( cl != m_clients.end() )
    {
        cl->second->sendMessage( &msg );        
        cl++;
    }
    if( dodelete ) delete itr->second; 
    m_localboxes.erase( itr );
    return false;
}


//=============================================================================
// Function Name: ConnectClient
///     This function is called when a POServer gets a new connection.  The
///     connection is connected by sending it a list of POBoxes in this 
///     post office.
/// \param client - the new client
//=============================================================================
void PostOffice::connectClient( POClient* client )
{
    Tracer::stat(7, "PostOffice::connectClient(...)");
    POBoxListMsg msg( POBoxListMsg::NEWBOXES );
    msg.setSource( POAddress( m_poname, "PostOffice" ) );
    msg.setDestination( POAddress( "", "PostOffice" ) );
    
    map<string, POBox*>::iterator itr = m_localboxes.begin();
    while( itr != m_localboxes.end() )
    {
        msg.addBox( itr->first );
        itr++;
    }

    client->sendMessage( &msg );
    
    //=========================================================
    // We will add the client to our list once it has been
    // connected and we have received the first POBoxListMsg
    // containing the clients name
    //=========================================================
    connect( client, SIGNAL( connectionEstablished( POClient* ) ),
             this,   SLOT  ( addConnectedClient( POClient* ) ) );
    connect( client, SIGNAL( connectionLost( POClient* ) ),
             this, SLOT( closeConnection( POClient* ) ) );
}   

//=============================================================================
// Function Name: AddConnectedClient
///     Takes a connected client and adds it to the client list, m_clients
/// \param client - the client to add
//=============================================================================
void PostOffice::addConnectedClient( POClient* client )
{
    Tracer::debug( 5, "PostOffice::addConnectedClient(...)" );
    map<string, POClient*>::iterator itr = m_clients.find( client->getClientName() );
    if( itr != m_clients.end() )
    {
        Tracer::error("DUPLICATE CLIENT NAME: %s", client->getClientName().c_str() );
//        POErrorMsg error( "DUPLICATE CLIENT NAME" );
//        error.setSource( POAddress( name, "PostOffice" ) );
//        error.setDestination( POAddress( "", "PostOffice" ) );
        delete client;
        return; 
    }
    m_clients[client->getClientName()] = client;
    connect( client, SIGNAL( receivedMessage( const POMessage* ) ),
             this, SLOT( deliverMessage( const POMessage* ) ) );

    emit newClient( m_poname, client->getClientName() );
}


//=============================================================================
// Function Name: SendMessage
///     Identifies where the message needs to go and sends it.
///     Assumes that the to/from fields have been filled out by the POBox
/// \param msg - a pointer to the message
/// \param throw_exception - true if you want to throw an exception instead
///                         of returning false.
/// \throws POErrorMsg if the destination post office cannot be found and
///                    and throw_exception is true
/// \return true on success, false on failure
//=============================================================================
bool PostOffice::sendMessage( const POMessage* msg, bool throw_exception /*=false*/ )
{
    POAddress adr = msg->getDestination();
    if( adr.getPO() == m_poname )
    {
        deliverMessage( msg );
        return true;
    }
    map<string, PostOffice*>::iterator pitr = m_postoffices.find( adr.getPO() );
    if( pitr != m_postoffices.end() ) 
    {
        pitr->second->deliverMessage( msg );
        return true; 
    }
    map<string, POClient*>::iterator itr = m_clients.find( adr.getPO() );
    if( itr == m_clients.end() )
    {
        if( throw_exception )
            throw new POErrorMsg("Post Office \"%s\" not found", adr.getPO().c_str() );
        else
            return false;
    }
    itr->second->sendMessage( adr, msg );
    return true;
}
//=============================================================================
// Function Name: sendMessage
///     Sends a message to the specified address from the specified pobox in
///     this PostOffice to the specified addres and returns true if the destination
///     exists
///     Throws a POErrorMsg* if the destination can't be found
/// \param msg - the message to send
/// \param from - the local POBox name
/// \param to - the destination address
//=============================================================================
void PostOffice::sendMessage( POMessage* msg, const string& from,
                              const POAddress& to )
{
    msg->setSource( POAddress( m_poname, from ) );
    msg->setDestination( to );
    sendMessage( msg );
}

bool PostOffice::deliverMessage( const POMessage* msg )
{
    POAddress adr = msg->getDestination();
    map<string, POBox*>::iterator itr = m_localboxes.find( adr.getBox() );
    if( itr == m_localboxes.end() )
    {
        Tracer::error( 5,"POBox %s not found at Post Office %s",
                       adr.getBox().c_str(), adr.getPO().c_str() );
        return false;
    }
    itr->second->addMessage( msg->clone() );
    return true;
}



//=============================================================================
// Function Name: ConnectTo
///     Creates a new client and attempts to connect to the specified server
/// \param host - the hostname/ip of the server
/// \param port - the port to use
/// \return The name of the remote post office
//=============================================================================
string PostOffice::connectTo( const string& host, int port, bool block )
{
    Tracer::stat( 5, "PostOffice::connectTo( %s:%d )", host.c_str(), port );
    POClient* client = new POClient( host, port );
    connectClient( client );
    if( !block ) return "NOT BLOCKING";

    DTime t;
    t.start();
    while( t.elapsed() < DTime( 3, 0 ) && 
           client->getState() != POClient::CONNECTED )
    {
        qApp->processEvents();
    }
    if( client->getState() != POClient::CONNECTED )
    {
        Tracer::warn( 3, "PostOffice::connectTo( %s:%d ) timed out", host.c_str(), port );
    }
    fprintf( stderr, "Before client->getClientName()\n\n" );
    return client->getClientName();
}

//=============================================================================
// Function Name: ServicePort
///     Creates a new POServer that listens for connections on the specified
///     port
/// \param  port - the port to listen on
//=============================================================================
bool PostOffice::servicePort( int port )
{
    cout<<"PostOffice::servicePort("<<port<<" )\n";
    Tracer::stat( 2, "Opening port %d", port );
    map<int, POServer*>::iterator itr = m_servers.find(port);
    if( itr != m_servers.end() )
    {
        Tracer::warn( 5, "Port %d is already open for connections.", port );
        return true;
    }
    QHostAddress hadr;
    hadr.setAddress("127.0.0.1");
    POServer* server = new POServer( hadr, port );
    if( !server->ok() )
    {
        Tracer::warn("Failed to open port %d", port );
        delete server;
        return false;
    }
    connect( server, SIGNAL( newClient( POClient* ) ),
             this, SLOT( connectClient( POClient* ) ) );
    m_servers[port] = server;
    return true;
}

//=============================================================================
// Function Name: ClosePort
///     Closes the server on a given port
/// \param port - the port to close
//=============================================================================
void PostOffice::closePort( int port )
{
    map<int, POServer*>::iterator itr = m_servers.find(port);
    if( itr == m_servers.end() )
    {
        Tracer::warn("PostOffice::closePort( %d ) - Post office wasn't servicing port", port);
        return;
    }
    delete itr->second;
    m_servers.erase( itr );
}


//=============================================================================
// Function Name: CloseConnection
///     Closes a connection and removes it from the client list
/// \param client - the client to close
//=============================================================================
void PostOffice::closeConnection( POClient* client )
{
    map<string, POClient*>::iterator itr = 
                m_clients.find( client->getClientName() );
    if( itr == m_clients.end() )
    {
        Tracer::warn(3, "PostOffice::CloseConnection() - Unknown Client" );
        return;
    }
    if( client != itr->second )
    {
        Tracer::error("PostOffice::CloseConnection(%s) - client pointer mismatch\n"
                      "\texpected %d, but found %d which has the name: %s",
                       client->getClientName().c_str(),
                       (int)itr->second, (int)client,
                       itr->first.c_str() );
        exit(1);
        return;
    }
    emit clientClosed( getName(), client->getClientName() );
    delete client;
    m_clients.erase(itr);
}



//=============================================================================
// Function Name: popMessage
///     This function allows you to pop a message from the specified POBox
/// \param box_name - the name of the POBox to pop the message from
/// \return a pointer to the oldest message in the box, NULL if no message found
//=============================================================================
const POMessage* PostOffice::popMessage( const string& box_name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(box_name);
    if( itr == m_localboxes.end() )
    {
        Tracer::warn( 5, "PostOffice::popMessage( %s ) - Unknown POBox" );
        return NULL;
    }
    return itr->second->popMessage();
}


//=============================================================================
// Function Name: hasMessage
///     Checks a POBox for messages
/// \param box_name - the name of the box to check
/// \return true if there are messages, otherwise false
//=============================================================================
bool PostOffice::hasMessage( const string& box_name )
{
    map<string, POBox*>::iterator itr = m_localboxes.find(box_name);
    if( itr == m_localboxes.end() )
    {
        Tracer::warn( 5, "PostOffice::popMessage( %s ) - Unknown POBox" );
        return false;
    }
    return itr->second->hasMessages();
}




//=============================================================================
// Function Name: getAllPostOffices
///     This function returns a pointer to all post offices running within
///     the current process
/// \return a map of Post Office Name to PostOffice pointer.
//=============================================================================
map<string, PostOffice*> PostOffice::getAllPostOffices()
{
    return m_postoffices;
}

PostOffice* PostOffice::getPostOffice( const string& po )
{
    cout<<"getPostOffice("<<po<<")\n";
    return m_postoffices[po];
}

//=============================================================================
// Function Name: getAllClients
/// \return   Returns a map of all connected clients, this should not be used by
///             non-api developers.
//=============================================================================
map<string, POClient*> PostOffice::getAllClients()const
{
    return m_clients;
}

POClient*  PostOffice::getClient( const string& cname )const
{
    map<string, POClient*>::const_iterator itr = m_clients.find( cname );
    if( itr != m_clients.end() )
        return itr->second;
    Tracer::warn( 5, "PostOffice::getClient( %s ) - Unknown client name", cname.c_str() );
    return NULL;
}


//=============================================================================
// Function Name: registerGroup
///     Creates a new group with the specified name
/// \throws POErrorMsg* if the group name is already in use by another POBox
/// \param name - the name of the group
/// \throws POErrorMsg if the group could not be created and throw_exception is true
/// \return a pointer to the group, null on error
//=============================================================================
POGroup* PostOffice::registerGroup( const string& name, bool throw_exception )
{
    if( m_localboxes.find(name) != m_localboxes.end() ) 
    {
        if( throw_exception )
        {
            throw new POErrorMsg("PostOffice::registerGroup( \"%s\" )\n"
                                 "\tThe name \"%s\" is already in use.",
                                 name.c_str(), name.c_str());
        }
        else
            return false;
    }
    POGroup* gr = new POGroup( name, this );
    m_groups[name] = gr;
    return gr;
}


//=============================================================================
// Function Name: closeGroup
///     Removes a group
/// \param name - the name of the group to close
/// \return true if the group was closed
//=============================================================================
bool PostOffice::closeGroup( const string& name )
{
    map<string, POGroup*>::iterator gr = m_groups.find(name);
    if( gr == m_groups.end() ) return false;
    delete gr->second;
    m_groups.erase( gr );
    return true;
}


//=============================================================================
// Function Name: getGroup
///     Returns a pointer to the group based on name
///     Throws an error if the group doesn't exist and create == false
/// \param name - the name of the group to get
/// \param create - true if the group should be created if it doesn't exist
/// \return a pointer to the group
//=============================================================================
POGroup* PostOffice::getGroup( const string& name, bool create, bool throw_exception )
{
   map<string, POGroup*>::iterator itr = m_groups.find(name);
   if( itr != m_groups.end() ) return itr->second;
   if( create ) return registerGroup(name);
    
   if( !throw_exception ) return NULL;

   throw new POErrorMsg( "PostOffice::getGroup(\"%s\", false)\n"
                         "\tGroup \"%s\" doesn't exist!\n",
                         name.c_str(), name.c_str() );
}


//=============================================================================
// Function Name: addAddressToGroup
///     Adds the specified address to the group.  This is useful for 
///     aggregating groups.  Be careful, the group may not be able to send to
///     the specified address.
/// \param adr - the address to add to the group
/// \param groupadr - the address of the group
/// \throws a POErrorMsg if the group address cannot be resolved 
//=============================================================================
void PostOffice::addAddressToGroup( const POAddress& adr, 
                                    const POAddress& groupadr )
{
    POGroupMsg addmsg( adr, POGroupMsg::ADD );
    try{
        sendMessage( &addmsg, "PostOffice", groupadr );
    }catch( POErrorMsg* err )
    {
        string errstr(err->getText());
        delete err;
        throw new POErrorMsg( "%s\nPostOffice::addAddressToGroup( %s.%s, %s.%s )\n"
                               "\tNo such address: %s.%s", 
                               errstr.c_str(),
                               adr.getPO().c_str(), adr.getBox().c_str(),
                               groupadr.getPO().c_str(), 
                               groupadr.getBox().c_str(),
                               groupadr.getPO().c_str(), 
                               groupadr.getBox().c_str()
                               );
    }
}

//=============================================================================
// Function Name: addAddressToGroup
///     Adds the specified address to the group.  This is useful for 
///     aggregating groups.  Be careful, the group may not be able to send to
///     the specified address.
/// \param adr - the address to add to the group
/// \param groupadr - the address of the group
/// \throws a POErrorMsg if the group address cannot be resolved 
//=============================================================================
void PostOffice::removeAddressFromGroup( const POAddress& adr, 
                                    const POAddress& groupadr )
{
    POGroupMsg rmmsg( adr, POGroupMsg::REMOVE );
    try{
        sendMessage( &rmmsg, "PostOffice", groupadr );
    }catch( POErrorMsg* err )
    {
        string errstr(err->getText());
        delete err;
        throw new POErrorMsg( "%s\nPostOffice::removeAddressFromGroup( %s.%s, %s.%s )\n"
                               "\tNo such address: %s.%s", 
                               errstr.c_str(),
                               adr.getPO().c_str(), adr.getBox().c_str(),
                               groupadr.getPO().c_str(), 
                               groupadr.getBox().c_str(),
                               groupadr.getPO().c_str(), 
                               groupadr.getBox().c_str()
                               );
    }
}
