#ifdef WIN32
#pragma warning(disable:4786)
#endif


#define IOSTREAMH
#include"Tracer.h"
#include<iostream>
#include<stdio.h>
#include<stdarg.h>
#include<vector>
#include<string>
using namespace std;
bool Tracer::m_initialized = false;
//=============================================
// This function is called on load, checks
// the env vars to see what to print
//=============================================
void Tracer::init()
{
    cout<<"Tracer::init!\n";
    if( m_initialized ) return;
    cout<<"HELLO?\n";
    //==============================================
    // Add Search Directorys from DGL_PLUGIN_PATH
    //==============================================
    char* spewargs = getenv("TRACER_SPEW");
    if( !spewargs ) 
        spewargs = ":";
    Tracer::debug("TRACER_SPEW = %s", spewargs );
    string spewarg(spewargs);
    spewarg+=':';
    string spewvar;
    for( unsigned int i = 0; i < spewarg.size(); i++ )
    {
         if( spewarg[i] != ':' && i != spewarg.size()-1)
             spewvar += spewarg[i];
         else if( spewvar.size() > 0 )
         {
             if( spewvar == "TRACE" )
                 Tracer::printTrace(true);
             else if( spewvar == "DEBUG" )
                 Tracer::printDebug(8);
             else if( spewvar == "STAT" )
                 Tracer::printStatus(8);
             else if( spewvar == "WARN" )
                 Tracer::printWarn(5);
             else if( spewvar == "NOERROR" )
                 Tracer::printError(0);
             else
                cout<<"UNKNOWN SPEW VAR: "<<spewvar<<endl;
             spewvar = string();
         }
    }
    m_initialized = true;
}


#ifndef WIN32
#define END     "\033[0m"

#define BLD     "\033[5m"
#define UND     "\033[4m"
#define REV     "\033[7m"

#define RED     "\033[31m"
#define GRN     "\033[32m"
#define YEL     "\033[33m"
#define BLU     "\033[34m"
#define VIL     "\033[35m"
#define TUR     "\033[36m"

#define RRED     "\033[41m"
#define RGRN     "\033[42m"
#define RYEL     "\033[43m"
#define RBLU     "\033[44m"
#define RVIL     "\033[45m"
#define RTUR     "\033[46m"
#else
#define END      ""

#define BLD      ""
#define UND      ""
#define REV      ""

#define RED      ""
#define GRN      ""
#define YEL      ""
#define BLU      ""
#define VIL      ""
#define TUR      ""

#define RRED     ""
#define RGRN     ""
#define RYEL     ""
#define RBLU     ""
#define RVIL     ""
#define RTUR     ""
#endif

//=============================================================================
// initialize The Static Member Vars
//=============================================================================
int  Tracer::m_level  = 0;
int  Tracer::m_indent = 2;
bool Tracer::m_printtrace = 0;
int  Tracer::m_printdebug = 0;
int  Tracer::m_printerror = 10;
int  Tracer::m_printwarn = 0;
int  Tracer::m_printstat = 0;
vector< pair<string, bool> > Tracer::m_stack;

Tracer::Tracer( const string& fname, bool print )
{
#ifdef WIN32
//===========================================
// Windows loads floating point library only
// when necessary and does not recognize the
// necessity when embedded in format string
// - this line forces it to load
//===========================================
	float tmp = 1.0f;
#endif

    enter( fname, print );
}
Tracer::~Tracer()
{
    leave();
}   

//=============================================================================
//  Function Name: Tracer::enter
///      enters a new level, printing the function name on the current level
///   and then printing everything else 1 indent further in until leave is 
///   called
/// 
/// \param function - the function that you are entering
//=============================================================================
void Tracer::enter( const string& function, bool print )
{
    if( !m_printtrace ) return;
    m_stack.push_back( pair<string, bool>(function, print) );
    if( !print ) return;
    m_level++;
    printIndent();
    cout<<"enter: "<<function<<endl;
}

//=============================================================================
//  Function Name: Tracer::leave
///     Pops one item off the call stack 
//=============================================================================
void Tracer::leave()
{
	if( m_stack.size() == 0 ) return;
    if( !m_stack.back().second ) 
    {
        m_stack.pop_back();
        return;
    }
    printIndent();
    m_level--;
    cout<<"Exit: "<<m_stack[m_stack.size()-1].first<<endl;
    m_stack.pop_back();
}


//=============================================================================
//  Function Name: Tracer::Debug
///      prints text at the current indent level
/// 
/// \param txt - the debug text to print
//=============================================================================
void Tracer::debug( const char* format, ... )
{
//    if( !m_printdebug ) return;
    printIndent();
    cout<<GRN<<"[DEBUG] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}

//=============================================================================
//  Function Name: Tracer::stat
///      prints text at the current indent level
/// 
/// \param txt - the debug text to print
//=============================================================================
void Tracer::stat( const char* format, ... )
{
    if( !m_printstat ) return;
    printIndent();
    cout<<BLU<<"[STATUS] "<<END;
    va_list ap;
    va_start(ap, format);
	vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}
//=============================================================================
// Function Name: error
///     prints an error to std error
/// \param format - the format of the error message
/// \param arguments for the format
//=============================================================================
void Tracer::error( const char* format, ... )
{
    if( !m_printerror ) return;
    printIndent();
    cerr<<RED<<"[ERROR] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stderr, format, ap );
    va_end(ap);
    cerr<<endl;
}

void Tracer::warn( const char* format, ... )
{
    if( !m_printwarn ) return;
    printIndent();
    cout<<YEL<<"[WARNING] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}

//=============================================================================
//  Function Name: Tracer::Debug
///      prints text at the current indent level
/// 
/// \param txt - the debug text to print
//=============================================================================
void Tracer::debug( int level, const char* format, ... )
{
    if( level > m_printdebug  ) return; // the level is greater than what we
                                        // are printing
    printIndent();
    cout<<GRN<<"[DEBUG ("<<level<<")] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}

//=============================================================================
//  Function Name: Tracer::stat
///      prints text at the current indent level
/// 
/// \param txt - the debug text to print
//=============================================================================
void Tracer::stat( int level, const char* format, ... )
{
    if( level > m_printstat ) return;
    printIndent();
    cout<<BLU<<"[STATUS ("<<level<<")] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}
//=============================================================================
// Function Name: error
///     prints an error to std error
/// \param format - the format of the error message
/// \param arguments for the format
//=============================================================================
void Tracer::error( int level, const char* format, ... )
{
    if( level > m_printerror ) return;
    printIndent();
    cerr<<RED<<"[ERROR ("<<level<<")] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stderr, format, ap );
    va_end(ap);
    cerr<<endl;
}

void Tracer::warn( int level, const char* format, ... )
{
    if( level > m_printwarn ) return;
    printIndent();
    cout<<YEL<<"[WARNING ("<<level<<")] "<<END;
    va_list ap;
    va_start(ap, format);
    vfprintf( stdout, format, ap );
    va_end(ap);
    cout<<endl;
}

//=============================================================================
//  Function Name: Tracer::printIndent
///      prints the white space at the beginning of the line
/// 
/// \return 
//=============================================================================
void Tracer::printIndent()
{
    for( int i = 0; i < m_level * m_indent; i++ )
        cout<<" ";
}
