//=============================================================================
// Tracer.h
//  - This class provides an elegant way to manage printing the stack as 
//  different functions are called and indents their start and stop tags to
//  make seeing recusion easy
//  
//  Created 10/5/02 By Daniel Larimer
//=============================================================================
#ifndef _TRACER_H_
#define _TRACER_H_
#include<string>
#include<vector>

#ifdef WIN32
#ifdef TRACER_EXPORTS
#define PEXPR_API __declspec(dllexport)
#else
#define PEXPR_API __declspec(dllimport)
#endif /* TRACER_EXPORTS */
#else
#define PEXPR_API
#endif /* WIN32 */

using namespace std;

class PEXPR_API Tracer 
{
    public:
        static void init();
        Tracer( const string& enter, bool print = true );
        ~Tracer();
        static void enter( const string& function, bool print = true );
        static void debug  ( const char* format, ... );
        static void error  ( const char* error, ... );
        static void warn   ( const char* warn, ...  );
        static void stat   ( const char* stat, ...  );
        static void debug  ( int level, const char* format, ... );
        static void error  ( int level, const char* error, ... );
        static void warn   ( int level, const char* warn, ...  );
        static void stat   ( int level, const char* stat, ...  );

        template<class ttype>
        static ttype& leave( ttype& rtn_value );
        static void leave();
		static int leave( int x ) { leave(); return x; }        
        static void printTrace ( bool state ) { m_printtrace = state; }
        static void printDebug ( int state )  { m_printdebug = state; }
        static void printError ( int state ) { m_printerror = state; }
        static void printWarn  ( int state ) { m_printwarn  = state; }
        static void printStatus( int state ) { m_printstat  = state; }
        static void setIndentSpaces( int num ) { m_indent = num;     }
    private:
        static void printIndent();

        static bool m_initialized;
        static vector< pair<string, bool> > m_stack;
        static bool m_printtrace;
        static int m_printdebug;
        static int m_printerror;
        static int m_printwarn;
        static int m_printstat;
        static int  m_level;
        static int  m_indent;
};

//=============================================================================
//  Function Name: Tracer::leave
///      leaves a level, printing the function name on the level it was called from
///   and then decreasing the indent
///   called
/// 
/// \param
//=============================================================================
template<class ttype>
ttype& Tracer::leave(ttype& rtn_value)
{
    if( !m_printtrace || m_stack.size() == 0 ) return rtn_value;
    if( !m_stack.back().second ) 
    {
        m_stack.pop_back();
        return rtn_value;
    }
    printIndent();
    m_level--;
//    cout<<"Exit: "<<m_stack[m_stack.size()-1].first<<endl;
    m_stack.pop_back();
    return rtn_value;
}
#endif
